#!/usr/bin/env sage

# The public key to attack.

pk = bytes.fromhex(sys.argv[1])

# Expected lengths (computed by estimate.sage).

estimates = (20.5290718248803, 62.0870996502688, 12748.7375083251, 69.7535812754563)

################################################################

load('emle-bytes.sage')
load('emle-stuff.sage')

################################################################

# Attack algorithms.

def solve_x_k0(h):

    norms = estimates[:3]
    print('average norms:', norms)
    A, B, C = (ceil(max(norms) / l) for l in norms)
    big = ceil(2^log(max(norms)/sqrt(n),2))  # or something

    short = sqrt(3*max(norms)^2 + big^2) * 1.5#ish
    print('target vector bound:', short)

    hprime = vector(h) - conv(GG[0], GG[1]) - k0base*pp[0]
    conv_gprime = matrix(conv(gprime, (ZZ^n).gen(i)) for i in range(n))

    mat = block_matrix(Zmod(pp[2]), [[conv_gprime], [pp[0]], [pp[1]]])
    sol = mat.solve_left(hprime)
    ker = mat.left_kernel_matrix()

    lattice = block_matrix(ZZ, [
            [ker,               0],
            [matrix([sol]),     1],
            [pp[2],             0],
        ])
    print('dimensions:', lattice.dimensions())

    print('weights:', A, B, C)
    weights = diagonal_matrix([A]*n + [B]*n + [C]*n + [big])


    # First run LLL and chop off zeroes.

    import time; t0 = time.time()
    print(f'\x1b[90mdoing LLL...\x1b[0m', end=' ', flush=True)
    reduced = (lattice * weights).LLL() * ~weights
    print(f'\x1b[90mdone; took {ceil(time.time()-t0)} seconds\x1b[0m')
    reduced = matrix(ZZ, filter(bool, reduced))
    print(f'shortest: \x1b[35m{(reduced[0]*weights).norm().n()}\x1b[0m {reduced[0]}')
    sys.stdout.flush()
    assert reduced.nrows() == reduced.ncols()


    # Run some fpylll BKZ to get started.

    for bsz in range(5,21,5):
        import time; t0 = time.time()
        print(f'\x1b[90mdoing BKZ-{bsz}...\x1b[0m', end=' ', flush=True)
        reduced = ((reduced * weights).BKZ(block_size=bsz) * ~weights).change_ring(ZZ)
        print(f'\x1b[90mdone; took {ceil(time.time()-t0)} seconds\x1b[0m')
        print(f'shortest: \x1b[35m{(reduced[0]*weights).norm().n()}\x1b[0m {reduced[0]}')
        sys.stdout.flush()


    # Now, run G6K BKZ tours progressively until we find a short vector.

    from fpylll import IntegerMatrix, GSO, LLL
    from fpylll.tools.bkz_stats import dummy_tracer
    from g6k import Siever
    from g6k.algorithms.bkz import pump_n_jump_bkz_tour

    M = LLL.reduction(IntegerMatrix.from_matrix(reduced * weights))
    g6k = Siever(M)
    assert g6k.max_sieving_dim >= reduced.ncols()

    for bsz in range(21, 91):
        import time; t0 = time.time()
        print(f'\x1b[90mdoing G6K-{bsz}...\x1b[0m', end=' ', flush=True)
        pump_n_jump_bkz_tour(g6k, dummy_tracer, bsz, pump_params={"down_sieve": True})
        print(f'\x1b[90mdone; took {ceil(time.time()-t0)} seconds\x1b[0m')
        print(f'shortest: \x1b[35m{vector(M[0]).norm().n()}\x1b[0m {vector(M[0])*~weights}')
        sys.stdout.flush()
        if vector(M[0]).norm() < short:
            break
    else:
        print('\x1b[31mFAILED\x1b[0m')
        exit(1)
    print(f'last block size: \x1b[36m{bsz}\x1b[0m')

    reduced = (matrix(list(map(list, M))) * ~weights).change_ring(ZZ)


    for row in reduced:
        if row[-1] < 0:
            row = -row
#        if row[-1] != 1:
#            print(f'\x1b[34m{(row * weights).norm().n()}\x1b[0m')
#            print(row)
        if row[-1] == 1:
            print(f'shortest vector found: \x1b[36m{(row * weights).norm().n()}\x1b[0m')
            print(row)
            my_x = row[:n]
            my_k0 = k0base + row[n:2*n]
            return my_x, my_k0

    assert False

def solve_k1_k2(h, x, k0):

    norms = estimates[2:]
    print('average norms:', norms)
    C, D = (ceil(max(norms) / l) for l in norms)
    big = ceil(2^log(max(norms)/sqrt(n),2))  # or something

    hprime = h - conv(GG[0],GG[1]) - conv(gprime,x) - k0*pp[0]

    mat = block_matrix(ZZ, [[pp[1]], [pp[2]], [-matrix([hprime])]])
    ker = mat.left_kernel_matrix()

    lattice = ker
    print('dimensions:', lattice.dimensions())

    print('weights:', C, D)
    weights = diagonal_matrix([C]*n + [D]*n + [big])

    reduced = (lattice * weights).BKZ(block_size=20) * ~weights
    reduced = matrix(ZZ, filter(bool, reduced))

    vec = reduced[0]
    if vec[-1] < 0:
        vec = -vec
    assert vec[-1] == 1

    k1 = vec[:n]
    k2 = vec[n:2*n]

    assert h - conv(GG[0],GG[1]) == conv(gprime,x) + k0*pp[0] + k1*pp[1] + k2*pp[2]

    return k1, k2

def reconstruct_F(x, k0, k1):
    F0 = conv(GG[0], x + GG[1]) + k0*pp[0]
    F1 = F0 + conv(GG[1], x) + k1*pp[1]
    return (F0, F1)

def solve(h):
    x, k0 = solve_x_k0(h)
    k1, k2 = solve_k1_k2(h, x, k0)
    F = reconstruct_F(x, k0, k1)
    return x, F

################################################################

# Solve it!

proof.all(False)    # faster

import time
start_time = time.time()

h1, h2 = decode_pk(pk)

x1, F1 = solve(h1)
print()
print(f'\x1b[32m{x1 = }\x1b[0m')
print()
x2, F2 = solve(h2)
print()
print(f'\x1b[32m{x2 = }\x1b[0m')
print()

# x_max == 4 for all parameter sets
assert all(abs(c) <= 4 for c in x1)
assert all(abs(c) <= 4 for c in x2)

################################################################

# Re-encode the recovered secret key.


print()
print()
print(f'\x1b[33mpk = {pk.hex().upper()}\x1b[0m')

sk = encode_sk((x1,x2,F1,F2), (h1,h2))
print()
print(f'\x1b[32msk = {sk.hex().upper()}\x1b[0m')
print()

print(f'\x1b[90mtotal time: {time.time() - start_time}\x1b[0m')
print()

